"""

EPP - effectful pythonic parsers, Core module.

The core module provides base classes and some essential parsers.

Parsing functions, or parsers, are just callables that take a State object as
the only argument and return another State object.

If a ParsingFailure exception is thrown by a parser, parsing stops with a
failure.
If a ParsingEnd exception is thrown by a parser, parsing ends prematurely, but
successfully.

"""


from collections import deque, namedtuple
import enum
import itertools as it


import epp.errors as error


#--------- base things ---------#


class State(namedtuple("State", "string effect left_start left_end parsed_start parsed_end")):
    """
    State objects represent current state of a parser chain (or an individual
    parser).

    State objects provide two views over the input string: 'left', which spans
    a substring between 'left_start' and 'left_end' and represents unparsed
    input left from the previous parser, and 'parsed', which spans a substring
    between 'parsed_start' and 'parsed_end' and represents a portion of input
    the previous parser has parsed. Windows may overlap, and usually
    'parsed_end' and 'left_start' are the same, but not always.

    A State object is immutable and has following fields:
    * string (str): the input the parser chain is supposed to parse.
    * effect ((value, state) -> value): if the chain is successful, this will
      be called in sequence with other effects from the chain to form the
      chain's output value.
    * left_start, left_end (int): see above about the 'left' window.
    * parsed_start, parser_end (int): see above about the 'parsed' window.

    State objects are just named tuples, so they support a very convenient
    '_replace' method. !Note!: to avoid duplicating effects accidentally,
    '_replace' treats lack of 'effect' in its arguments as 'effect=None'. So if
    you want to copy an effect from another parser, you have to do it
    explicitly.

    State objects' constructor takes the following arguments:
    1. string - the input.
    2. effect=None - the effect, transformation to be performed on success of
       the last parser.
    3. start=0 - will be translated into 'left_start'
    4. end=None - will be translated into 'left_end'. If set to None,
      'left_end' will be set to the length of the input.
    State objects created via this constructor have both 'parsed_start' and
    'parsed_end' set to 'left_start'.

    State objects have several properties:
    * left - returns a slice of input that's left to parse.
    * left_len - returns the length of the above slice without computing the
      slice itself.
    * parsed - returns a slice of input that's been parsed.
    * parsed_len - returns the length of the above slice, again without
      computing the slice.

    Finally, State objects have following public methods:
    * consume(how_many) - move 'how_many' characters from the left window into
      the parsed window. Raise ValueError if more input was consumed than left.
    * split(at) - split the State in two (and return them). The first keeps
      the input up to, but not including, 'at' as its 'left' window, the second
      gets the rest. Both have their 'parsed' windows reset to an empty string.
      The first gets 'effect' of the original, the second gets None.
    """

    __slots__ = []

    def __new__(cls, string, effect=None, start=0, end=None):
        if end is None:
            end = len(string)
        assert 0 <= start <= end <= len(string)
        return super().__new__(cls, string, effect, start, end, start, start)

    def _replace(self, **kwargs):
        if "effect" not in kwargs:
            return super()._replace(effect=None, **kwargs)
        return super()._replace(**kwargs)

    def consume(self, how_many):
        """
        Return a new State object with 'how_many' characters consumed and moved
        to the 'parsed' window.

        Raise ValueError if 'how_many' is negative or if consuming more
        characters than left in the 'left' window.
        """
        if how_many < 0:
            raise ValueError("Negative number of consumed characters")
        left_start = self.left_start + how_many
        parsed_start = self.left_start
        parsed_end = parsed_start + how_many
        if left_start > self.left_end:
            raise ValueError("Consumed more characters than fits in the 'left' window")
        return self._replace(left_start=left_start, parsed_start=parsed_start,
                             parsed_end=parsed_end)

    def split(self, at):
        """
        Split the State in two. The first one keeps a portion of input up to
        'at'th character (exclusive), the second one gets the rest. Both have
        'parsed' window reset to an empty string. First one gets the effect of
        the original, the second one gets None.
        """
        split_point = self.left_start + at
        first = self._replace(effect=self.effect,
                              left_end=split_point,
                              parsed_start=self.left_start,
                              parsed_end=self.left_start)
        second = self._replace(effect=None,
                               left_start=split_point,
                               parsed_start=split_point,
                               parsed_end=split_point)
        return first, second

    @property
    def left(self):
        """
        Return the portion of input the last parser hasn't consumed.
        """
        return self.string[self.left_start:self.left_end]

    @property
    def left_len(self):
        """
        Return the length of the portion of input the last parser hasn't
        consumed.
        """
        return self.left_end - self.left_start

    @property
    def parsed(self):
        """
        Return the string parsed by the last parser.
        """
        return self.string[self.parsed_start:self.parsed_end]

    @property
    def parsed_len(self):
        """
        Return the length of the string parsed by the last parser.
        """
        return self.parsed_end - self.parsed_start


class ParsingFailure(Exception):
    """
    An exception of this type should be thrown if parsing fails.
    
    The constructor takes three arguments: the state that has caused a parser
    to fail, the error message and an optional code which (in predefined
    parsers) is used to tell apart the reasons for the failure within the same
    parser.
    """

    def __init__(self, failed_state, text, code=0):
        super().__init__(text)
        self.code = code
        self.state = failed_state


class ParsingEnd(Exception):
    """
    An exception of this type should be thrown if parsing ends successfully,
    but early.
    """

    def __init__(self, state):
        super().__init__()
        self.state = state


class Lookahead(enum.Enum):
    """ Lookahead type. """
    GREEDY = enum.auto()
    RELUCTANT = enum.auto()


def parse(seed, state_or_string, parser, verbose=False):
    """
    Run a given parser on a given state object or a string, then apply combined
    chain or parser's effects to 'seed' and return a tuple
    (seed after effects, final state).

    On failure, return None unless 'verbose' is truthy, in which case return
    the ParsingFailure exception that has terminated the parsing process.
    """
    if isinstance(state_or_string, str):
        state = State(state_or_string)
    else:
        state = state_or_string
    try:
        after = parser(state)
        if after.effect is not None:
            return after.effect(seed, after), after
        return seed, after
    except ParsingFailure as failure:
        if verbose:
            return failure
        return None
    except ParsingEnd as end:
        if end.state.effect is not None:
            return end.state.effect(seed, end.state), end.state
        return seed, end.state


#--------- core parsers generators ---------#


def branch(funcs, save_iterator=True):
    """
    Create a parser that will try given parsers in order and return the state
    of the first successful one.

    If 'save_iterator' is truthy (the default), the parsers from the given
    iterator will be saved, allowing the resulting parser to be run several
    times without the danger of the 'funcs' being consumed on the first run.
    However, this increases memory consumption. If you're absolutely sure that
    your branch will be run only once, or if you already use a reusable
    iterable such as a list or a deque, you can pass False as 'save_iterator'
    to avoid memory overhead.
    """
    saved = deque()
    if save_iterator:
        funcs = iter(funcs)
    def branch_body(state):
        """ A tree of parsers. """
        saved_len = len(saved)
        for i, parser in enumerate(it.chain(saved, funcs)):
            if i >= saved_len and save_iterator:
                saved.append(parser)
            try:
                return parser(state)
            except ParsingEnd as end:
                return end.state
            except ParsingFailure:
                continue
        raise ParsingFailure(state, 
                             "All parsers in a branching point have failed",
                             error.BranchError.ALL_FAILED)
    return branch_body


def catch(parser, exception_types, on_thrown=None, on_not_thrown=None):
    """
    Return a parser that runs 'parser' and catches exceptions of any of types
    given by 'exception_types'.

    If any exception was caught, 'on_thrown' is called as follows:
    > on_thrown(original_state, caught_exception)
    and its return value (which should be a State object) replaces the original
    parser chain's state.

    If no exception was caught, 'on_not_thrown' is called:
    > on_not_thrown(parsers_state)
    and its return value replaces parser chain's state.

    Both 'on_thrown' and 'on_not_thrown' may be None, in this case no action is
    performed.

    Note that ParsingFailure and ParsingEnd exceptions are exempt from being
    caught in this manner.
    """
    exception_types = tuple(exception_types)
    def catch_body(state):
        """ Try to catch an exception thrown by another parser. """
        try:
            if on_not_thrown is None:
                return parser(state)
            return on_not_thrown(parser(state))
        except ParsingFailure as failure:
            raise failure
        except ParsingEnd as end:
            raise end
        except Exception as exc:
            if isinstance(exc, exception_types):
                if on_thrown is not None:
                    return on_thrown(state, exc)
                return state
            raise exc
    return catch_body


def chain(funcs, combine=True, stop_on_failure=False, all_or_nothing=True,
          save_iterator=True):
    """
    Create a parser that chains a given iterable of parsers together, using
    output of one parser as input for another.

    If 'combine' is truthy, resulting 'parsed' window will cover the input
    between starting point of the first parser and the ending point of the
    last one. Note that this might be different than what you'd get by
    concatenating together individual 'parsed' windows if some of the parsers
    performed unusual operations on their windows - like 'noconsume' does, for
    example.

    If 'stop_on_failure' is truthy, stop parsing instead of failing it when a
    parser in the chain raises a ParsingFailure exception. Note that this also
    includes parsers with lookahead, effectively disabling it.

    If 'all_or_nothing' is truthy (the default), a ParsingEnd exception thrown
    inside it will not cause the effects gathered so far to be registered and
    the part of the string parsed so far to be marked as 'parsed'. If the
    parameter is falsey, partial application of effects and parsers is
    possible. 'all_or_nothing' is suppressed if 'stop_on_failure' is truthy.

    If 'save_iterator' is truthy (the default), the elements of the supplied
    iterator will be saved for future reuse. This avoids a problem of the
    iterator being exausted after the first parser run, but leads to higher
    memory consumption. Disable it only if you're sure that the chain will only
    be used once, or if you've wrapped the iterator into some reusable
    iterable, like a list or a deque, or if you've used 'reuse_iter'.
    """
    return _Chain(funcs, combine, stop_on_failure, all_or_nothing, save_iterator)


def effect(eff):
    """
    Register an effect in the chain. The argument should be a callable of two
    arguments with signature
    (value, state) -> value,
    where 'value' is an arbitrary object being constructed by chain's effects,
    and 'state' is the state of the parser chain at the moment of effect's
    registration. Return 'value' can either be a new object, or a modified old
    one.  Alternatively, you can supply the value being modified through a
    nonlocal variable - like
    > values = deque()
    > eff = effect(lambda val, state: values.append(state.parsed)).
    Note that this would tie your effect to the place of invokation and make it
    not particularly reusable.
    """
    def effect_(state):
        """ Register an effect. """
        return state._replace(effect=eff)
    return effect_


def fail():
    """ Return a parser that always fails. """
    def fail_body(state):
        """ Fail immediately. """
        raise ParsingFailure(state, "'fail' parser has been reached", error.FailError.FAILED)
    return fail_body


def identity():
    """ Return a parser that passes state unchanged. """
    return lambda state: state._replace()


def lazy(generator, *args, **kwargs):
    """
    Make 'generator' lazy. It will only be called when it's time to actually
    parse a string. Useful for recursive parsers.
    """
    def lazy_body(state):
        """ Lazily create a parser. """
        parser = generator(*args, **kwargs)
        return parser(state)
    return lazy_body


def modify_error(parser, error_transformer):
    """
    Return a parser that will run 'parser' and, if it fails, modifies raised
    ParsingFailure exception by running it through 'error_transformer' (which
    should be a callable of a single argument, and should return an exception)
    and re-raising its return value.
    """
    def modify_error_msg_body(state):
        """ Modify error message. """
        try:
            return parser(state)
        except ParsingFailure as failure:
            raise error_transformer(failure)
    return copy_lookahead(parser, modify_error_msg_body)


def noconsume(parser):
    """ Return a version of 'parser' that doesn't consume input. """
    def noconsume_body(state):
        """ Parse without consuming input. """
        output = parser(state)
        return output._replace(effect=output.effect, left_start=state.left_start)
    return noconsume_body


def stop(discard=False):
    """
    Return a parser that stops parsing immediately.

    If 'discard' is truthy, truncate 'parsed' window, otherwise inherit it from
    the previous parser.
    """
    def stop_body(state):
        """ Stop parsing. """
        if discard:
            state = state._replace(parsed_start=state.left_start,
                                   parsed_end=state.left_start)
        raise ParsingEnd(state._replace())
    return stop_body


def subparse(seed, parser, absorber):
    """
    Create a parser that will run 'parser' (via 'parse') on the current input
    and given seed value, and then register an effect incorporating subchain's
    output value into the main chain using 'absorber', which should be a
    callable:
    (main_chain_value, main_chain_state, parser_value, parser_state) -> value

    Parser's output state will replace chain's state if the operation is
    succesful.

    If parser fails, so does 'subparse'.
    """
    def absorb_inner(state):
        """ Absorb results of another parser into the chain. """
        output = parse(seed, state, parser)
        if output is None:
            raise ParsingFailure(state, "Subparsing failed", error.SubparseError.FAILED)
        value, after = output
        after = after._replace(effect=lambda val, st: absorber(val, state, value, after))
        return after
    return absorb_inner


def test(testfn):
    """
    Return a parser that succeeds consuming no input if testfn(state) returns a
    truthy value, and fails otherwise.

    'parsed' window is truncated.
    """
    def test_body(state):
        """ State testing function. """
        if testfn(state):
            return state._replace(parsed_start=state.left_start, parsed_end=state.left_start)
        raise ParsingFailure(state,
                             f"Function {testfn} returned a falsey value on '{state.left[0:20]}'",
                             error.TestError.FAILED)
    return test_body


#--------- helper things ---------#


def copy_lookahead(from_parser, to):
    """
    Copy lookahead mode from 'from_parser' to 'to' and return the modified
    parser.
    """
    try:
        to.lookahead = from_parser.lookahead
    except AttributeError:
        pass
    return to


def get_lookahead(parser):
    """
    Return lookahead mode of the parser or None if it doesn't perform lookahead.
    """
    try:
        return parser.lookahead
    except AttributeError:
        return None


def greedy(parser):
    """ Return a greedy version of 'parser'. """
    try:
        if parser.lookahead is Lookahead.GREEDY:
            return parser
    except AttributeError:
        pass
    res = lambda state: parser(state)
    res.lookahead = Lookahead.GREEDY
    return res


def has_lookahead(parser):
    """ Return True if the parser has the ability to perform lookahead. """
    return hasattr(parser, "lookahead")


def is_greedy(parser):
    """ Return True if the parser is greedy, False otherwise. """
    try:
        return parser.lookahead is Lookahead.GREEDY
    except AttributeError:
        return False


def is_reluctant(parser):
    """ Return True if the parser is reluctant, False otherwise. """
    try:
        return parser.lookahead is Lookahead.RELUCTANT
    except AttributeError:
        return False


def no_lookahead(parser):
    """ Return True if the parser performs no lookahead. """
    return not hasattr(parser, "lookahead")


def reluctant(parser):
    """ Return a reluctant version of 'parser'. """
    try:
        if parser.lookahead is Lookahead.RELUCTANT:
            return parser
    except AttributeError:
        pass
    res = lambda state: parser(state)
    res.lookahead = Lookahead.RELUCTANT
    return res


def reuse_iter(generator, *args, **kwargs):
    """
    Make an iterable that will call 'generator(*args, **kwargs)' when iterated
    over and use the return value as an iterator.
    """
    class Res():
        """ A reusable iterator container. """
        def __iter__(self):
            return generator(*args, **kwargs)
    return Res()


#--------- private helper things ---------#


def _chain_effects(effect_points):
    """ Chain effects saved in 'states' together into a single effect. """
    def chained_effects(value, state):
        """ A chain of effects. """
        for (s, _) in effect_points:
            value = s.effect(value, s)
        return value
    return chained_effects


def _overrestricted(parser):
    """ Return True if a parser is maximally restricted. """
    # isinstance may not be idiomatic, but it's safer than relying on parsers
    # not having a particular method.
    if not isinstance(parser, _RestrictedParser):
        return True
    return parser.overrestricted()


def _reset(parser):
    """ Reset restrictions on a parser. """
    if not isinstance(parser, _RestrictedParser):
        return
    parser.reset()


def _reset_chain(parsers, start_at):
    """ Reset all restricted parsers to the right of 'start_at'. """
    length = len(parsers)
    for i in range(start_at + 1, length):
        _reset(parsers[i])


def _restrict(parser, state):
    """
    Return a restricted version of a parser.
    A no-op when used on a parser that performs no lookahead.
    """
    if get_lookahead(parser) is None:
        return parser
    return _RestrictedParser(parser, state)


def _restrict_more(parser):
    """
    Further restrict a parser. A no-op when used on a parser that performs no
    lookahead.
    """
    if not isinstance(parser, _RestrictedParser):
        return
    parser.restrict_more()


def _shift(parsers, from_pos):
    """
    Propagate restrictions' change from 'from_pos' to the left end of a parser
    chain.

    Return the index of the last parser affected or None if all restriction
    combinations were tried.
    """
    while from_pos >= 0:
        p = parsers[from_pos]
        _restrict_more(p)
        if _overrestricted(p):
            from_pos -= 1
            continue
        return from_pos
    return None


def _partial_parse(state, parser, at):
    """ Parse using only a portion of the input (namely, up to 'at'). """
    use, do_not = state.split(at)
    after = parser(use)
    after = after._replace(effect=after.effect, left_end=do_not.left_end)
    return after


def _try_chain(parsers, from_pos, effect_points):
    """
    Try to parse the state the first parser in the chain remembers.

    Return a tuple (state, index of the first parser to fail).
    In case of failure, 'state' will be None.
    """
    state = parsers[from_pos].state_before
    new_effect_points = deque()
    drop_effects_after = effect_points.find(lambda point: point[1] >= from_pos)
    i = from_pos
    for i in range(from_pos, len(parsers)):
        try:
            parsers[i].state_before = state
            state = parsers[i](state)
            if state.effect is not None:
                new_effect_points.append((state, i))
        except ParsingFailure:
            return (None, i)
        except ParsingEnd as end:
            if drop_effects_after is not None:
                effect_points.drop(len(effect_points) - drop_effects_after)
            effect_points.extend(new_effect_points)
            raise end
    if drop_effects_after is not None:
        effect_points.drop(len(effect_points) - drop_effects_after)
    effect_points.extend(new_effect_points)
    return state, i


class _CachedAppender():
    """
    A class that tries to combine efficient appending of deques and fast
    indexing of lists.
    """

    def __init__(self):
        self.changed = False
        self.deque = deque()
        self.list = []
        self.empty = True

    def __getitem__(self, index):
        if self.empty:
            raise IndexError("Indexing into an empty appender")
        if self.changed:
            self.update()
        return self.list[index]

    def __iter__(self):
        return iter(self.deque)

    def __setitem__(self, key, value):
        self.deque[key] = value
        self.list[key] = value

    def __len__(self):
        if self.changed:
            return len(self.deque)
        return len(self.list)

    def append(self, item):
        """ Add an element to the right side of the underlying deque. """
        self.deque.append(item)
        self.changed = True
        self.empty = False

    def drop(self, num):
        """ Drop 'num' elements from the right end. """
        for _ in range(num):
            self.deque.pop()
            self.changed = True
        try:
            _ = self.deque[0]
        except IndexError:
            self.empty = True

    def extend(self, iterable):
        """ Append all elements of an iterable. """
        for item in iterable:
            self.deque.append(item)
            self.changed = True
            self.empty = False

    def find(self, pred):
        """
        Return the index of an element such that 'pred(element)' returns True
        or None if no such element was found.
        """
        self.update()
        for i, val in enumerate(self.deque):
            if pred(val):
                return i
        return None

    def update(self):
        """ Syncronize the underlying list with the deque. """
        self.list = list(self.deque)
        self.changed = False


class _Chain():
    """
    A chain of parsers.
    """

    def __init__(self, funcs, combine, stop_on_failure, all_or_nothing, save_iterator):
        self.combine = combine
        self.stop_on_failure = stop_on_failure
        self.all_or_nothing = all_or_nothing
        if save_iterator:
            self.parsers = iter(funcs)
            self.saved_parsers = deque()
        else:
            self.parsers = funcs
            self.saved_parsers = None
        self.effect_points = None
        self.first_state = None
        self.lookahead_chain = None

    def __call__(self, state):
        self.reset(state)
        return self.parse()

    def prep_output_state(self, state, early):
        """ Prepare output state: combine 'parsed's and effects. """
        if early and self.all_or_nothing and not self.stop_on_failure:
            return self.first_state
        if self.combine:
            state = state._replace(parsed_start=self.first_state.left_start)
        state = state._replace(effect=_chain_effects(self.effect_points))
        return state

    def prep_end_exception(self, end, state):
        """
        Prepare ParsingEnd exception: attach a prepared output state to it.
        """
        end.state = self.prep_output_state(state, True)
        return end

    def reset(self, state):
        """ Reset the state of the chain. """
        self.effect_points = _CachedAppender()
        self.first_state = state
        self.lookahead_chain = None

    def start_backtracking(self, state, parsers):
        """ Start backtracking, then continue with normal parsing. """
        start_from = len(self.lookahead_chain) - 1
        while True:
            pos = _shift(self.lookahead_chain, start_from)
            if pos is None:
                raise ParsingFailure(state,
                                     "No combination of inputs allows successful parsing",
                                     error.ChainError.LOOKAHEAD_FAILED)
            _reset_chain(self.lookahead_chain, pos)
            after, failed = _try_chain(self.lookahead_chain, pos, self.effect_points)
            if after is None:
                start_from = failed
                continue
            state = after
            try:
                for parser in parsers:
                    state = self.parse_one(state, parser)
                return self.prep_output_state(state, False)
            except ParsingEnd as end:
                raise self.prep_end_exception(end, state)
            except ParsingFailure:
                start_from = len(self.lookahead_chain) - 1
                continue

    def parse(self):
        """ Try to parse the initial state. """
        try:
            if self.saved_parsers is not None:
                saved = self.saved_parsers.copy()
                num_saved = len(saved)
                parsers = it.chain(saved, self.parsers)
            else:
                num_saved = -1
                parsers = self.parsers
            state = self.first_state
            for i, parser in enumerate(parsers):
                if num_saved >= 0 and i >= num_saved:
                    self.saved_parsers.append(parser)
                state = self.parse_one(state, parser)
            return self.prep_output_state(state, False)
        except ParsingEnd as end:
            raise self.prep_end_exception(end, state)
        except ParsingFailure as failure:
            if self.stop_on_failure:
                return self.prep_output_state(state, True)
            if self.lookahead_chain is None:
                raise failure
            return self.start_backtracking(state, parsers)

    def parse_one(self, state, parser):
        """ Parse using a single parser. Return the resulting state. """
        pos = None
        if self.lookahead_chain is None and has_lookahead(parser):
            self.lookahead_chain = _CachedAppender()
        if self.lookahead_chain is not None:
            parser = _restrict(parser, state)
            pos = len(self.lookahead_chain)
            self.lookahead_chain.append(parser)
        after = parser(state)
        if after.effect is not None:
            self.effect_points.append((after, pos))
        return after


class _RestrictedParser():
    """ A parser that only operates on a restricted portion of input. """

    def __init__(self, parser, state):
        self.parser = parser
        self.lookahead = get_lookahead(parser)
        self.delta = 0
        self.state_before = state

    def __call__(self, state):
        self.state_before = state
        if self.lookahead is None:
            return self.parser(state)
        if self.lookahead is Lookahead.GREEDY:
            return _partial_parse(state, self.parser, state.left_len - self.delta)
        # is reluctant
        return _partial_parse(state, self.parser, self.delta)

    def overrestricted(self):
        """
        Return True if restrictions have reached their maximum - that is, if
        either allowed input portion is shrinked into an empty string, or has
        extended beyond the bounds of leftover input.
        """
        return self.delta > self.state_before.left_end - self.state_before.left_start

    def reset(self):
        """ Reset restrictions. """
        self.delta = 0
        self.state_before = None

    def restrict_more(self):
        """ Increase restriction level on the input. """
        self.delta += 1



"""

Extras module.

This module provides some convenience things to use with parsers, mostly data
structures that are useful in effects.

"""


from collections import deque


class SRDeque(deque):
    """
    Self-returning deque.

    Supports all the operations of the normal deques, but returns itself on
    most operations that otherwise return None, making it usable in effects.
    """

    def append(self, elem):
        super().append(elem)
        return self

    def appendleft(self, elem):
        super().append(elem)
        return self

    def clear(self):
        super().clear()
        return self

    def extend(self, iterable):
        super().extend(iterable)
        return self

    def extendleft(self, iterable):
        super().extendleft(iterable)
        return self

    def insert(self, index, elem):
        super().insert(index, elem)
        return self

    def remove(self, elem):
        super().remove(elem)
        return self

    def reverse(self):
        super().reverse()
        return self

    def rotate(self, n=1):
        super().rotate(n)
        return self


class SRDict(dict):
    """
    Self-returning dict.

    Supports all the operations of the normal dicts, but returns itself on
    most operations that otherwise return None, making it usable in effects.

    Also provides 'set' operation which can be used in effects instead of item
    assignment.
    """

    def clear(self):
        super().clear()
        return self

    def update(self, other):
        super().update(other)
        return self

    def set(self, **kwargs):
        """
        Set the given elements of the dict to corresponding values, then return
        itself.
        """
        return self.update(kwargs)


class SRList(list):
    """
    Self-returning list.

    Supports all the operations of the normal lists, but returns itself on most
    operations that otherwise return None, making it usable in effects.

    Also provides 'set' method to use instead of item and slice assigment.
    """

    def append(self, elem):
        super().append(elem)
        return self

    def clear(self):
        super().clear()
        return self

    def extend(self, iterable):
        super().extend(iterable)
        return self

    def insert(self, index, elem):
        super().insert(index, elem)
        return self

    def remove(self, elem):
        super().remove(elem)
        return self

    def reverse(self):
        super().reverse()
        return self

    def sort(self, key=None, reverse=False):
        super().sort(key=key, reverse=reverse)
        return self

    def set(self, index, value):
        """
        Set the element as 'index' to 'value', then return self.

        'index' may also be a slice object, in this case 'value' should be an
        iterable.
        """
        self[index] = value
        return self



"""
Error module.

Provides error codes for parsers in 'core' and 'parsers'.
"""

from enum import Enum, auto


#--------- error codes for core parsers ---------#


class BranchError(Enum):
    """ Error codes for 'branch' parsers. """
    ALL_FAILED = auto()


class ChainError(Enum):
    """ Error codes for 'chain' parsers. """
    LOOKAHEAD_FAILED = auto()


class FailError(Enum):
    """ Error codes for 'fail' parsers. """
    FAILED = auto()


class SubparseError(Enum):
    """ Error codes for 'subparse' parsers. """
    FAILED = auto()


class TestError(Enum):
    """ Error codes for 'test' parsers. """
    FAILED = auto()


#--------- error codes for single-character parsers ---------#


class AlnumError(Enum):
    """ Error codes for 'alnum' parsers. """
    EOI = auto()
    NON_ALNUM = auto()


class AlphaError(Enum):
    """ Error codes for 'alpha' parsers. """
    EOI = auto()
    NON_ALPHA = auto()


class AnyCharError(Enum):
    """ Error codes for 'any_char' parsers. """
    EOI = auto()


class CondCharError(Enum):
    """ Error codes for 'cond_char' parsers. """
    EOI = auto()
    DID_NOT_PASS = auto()


class DigitError(Enum):
    """ Error codes for 'digit' parsers. """
    EOI = auto()
    NOT_DIGIT = auto()


class HexDigitError(Enum):
    """ Error codes for 'hex_digit' parsers. """
    EOI = auto()
    NOT_DIGIT = auto()


class NewlineError(Enum):
    """ Error codes for 'newline' parsers. """
    EOI = auto()
    NOT_NEWLINE = auto()


class NonwhiteError(Enum):
    """ Error codes for 'nonwhite_char' parsers. """
    EOI = auto()
    WHITE = auto()


class WhiteCharError(Enum):
    """ Error codes for 'white_char' parsers. """
    EOI = auto()
    NON_WHITE = auto()
    NEWLINE = auto()


#--------- error codes for aggregates of single-character parsers ---------#


class AlnumWordError(Enum):
    """ Error codes for 'alnum_word' parsers. """
    EOI = auto()
    NON_ALNUM = auto()


class AlphaWordError(Enum):
    """ Error codes for 'alpha_word' parsers. """
    EOI = auto()
    NON_ALPHA = auto()


class AnyWordError(Enum):
    """ Error codes for 'any_word' parsers. """
    EOI = auto()
    WHITE = auto()


class HexIntError(Enum):
    """ Error codes for 'hex_int' parsers. """
    EOI = auto()
    NON_HEX_INT = auto()
    NO_PREFIX = auto()


class IntegerError(Enum):
    """ Error codes for 'integer' parsers. """
    EOI = auto()
    NON_INT = auto()


class LineError(Enum):
    """ Error codes for 'line' parsers. """
    EOI = auto()


class WhitespaceError(Enum):
    """ Error codes for 'whitespace' parsers. """
    EOI = auto()
    NOT_ENOUGH = auto()


#--------- error codes for various parsers ---------#


class BalancedError(Enum):
    """ Error codes for 'balanced' parsers. """
    DOESNT_START = auto()
    NO_PAIR = auto()


class EndOfInputError(Enum):
    """ Error codes for 'end_of_input' parsers. """
    NOT_END = auto()


class LiteralError(Enum):
    """ Error codes for 'literal' parsers. """
    SHORTER = auto()
    DOESNT_START = auto()


class MultiError(Enum):
    """ Error codes for 'multi' parsers. """
    ALL_FAILED = auto()


class RepeatWhileError(Enum):
    """ Error codes for 'repeat_while' parsers. """
    NOT_ENOUGH = auto()


class TakeError(Enum):
    """ Error codes for 'take' parsers. """
    NOT_ENOUGH = auto()



from setuptools import setup, find_packages

from os import path

def long_description():
    """ Read the descrption from README. """
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, "README.rst")) as f:
        return f.read()

setup(
    name="epparsers",
    version="0.3.0",
    package_dir={"": "src"},
    packages=find_packages("src"),

    test_suite="test",

    author="Michail Pevnev",
    author_email="mpevnev@gmail.com",
    description="Effectful pythonic parsers",
    long_description=long_description(),
    license="LGPL-3",
    keywords="parser parsers",
    url="https://github.com/mpevnev/epp",

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)"
        ]
)



"""

Parsers module.

This module provides actually useful parsers, as opposed to the bookkeeping
ones in the 'core' module.

"""

from collections import deque
import itertools as itools

import epp.core as core
import epp.errors as error


#--------- single-character parsers ---------#


def alnum(ascii_only=False):
    """
    Return a parser that will match a single alphanumeric character.

    If 'ascii_only' is truthy, match only ASCII alphanumeric characters
    ([a-zA-Z0-9]), not whatever makes .isalnum() return True.
    """
    def alnum_body(state):
        """ Match an alphanumeric character. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected an alphanumeric character, got the end of input",
                error.AlnumError.EOI)
        if ascii_only:
            if 'a' <= char <= 'z' or 'A' <= char <= 'Z' or '0' <= char <= '9':
                return state.consume(1)
            raise core.ParsingFailure(
                state,
                f"Expected an alphanumeric character, got '{char}'",
                error.AlnumError.NON_ALNUM)
        if char.isalnum():
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected an alphanumeric character, got '{char}'",
            error.AlnumError.NON_ALNUM)
    return alnum_body


def alpha(ascii_only=False):
    """
    Return a parser that will match a single alphabetic character.

    If 'ascii_only' is truthy, match only ASCII alphabetic characters, not
    everything for which .isalpha() returns True.
    """
    def alpha_body(state):
        """ Match an alphabetic character. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected an alphabetic character, got the end of input",
                error.AlphaError.EOI)
        if ascii_only:
            if 'a' <= char <= 'z' or 'A' <= char <= 'Z':
                return state.consume(1)
            raise core.ParsingFailure(
                state,
                f"Expected an alphabetic character, got '{char}'",
                error.AlphaError.NON_ALPHA)
        if char.isalpha():
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected an alphabetic character, got '{char}'",
            error.AlphaError.NON_ALPHA)
    return alpha_body


def any_char():
    """ Return a parser that would match any character. """
    def any_char_body(state):
        """ Match a single character. """
        try:
            _ = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a character, got the end of input",
                error.AnyCharError.EOI)
        return state.consume(1)
    return any_char_body


def cond_char(condition):
    """
    Return a parser that will match a character such that 'condition(char)' is
    truthy.

    Raise ValueError if 'condition' is not callable.
    """
    if not callable(condition):
        raise ValueError(f"{condition} is not callable")
    def cond_char_body(state):
        """ Match a character that passes a conditional check. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a character, got the end of input",
                error.CondCharError.EOI)
        if condition(char):
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"{char} did not pass the {condition} test",
            error.CondCharError.DID_NOT_PASS)
    return cond_char_body


def digit():
    """
    Return a parser that would match a single decimal digit.
    """
    def digit_body(state):
        """ Parse a single decimal digit. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a digit, got the end of input",
                error.DigitError.EOI)
        if '0' <= char <= '9':
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected a digit, got '{char}'",
            error.DigitError.NOT_DIGIT)
    return digit_body


def hex_digit():
    """
    Return a parser that matches a single hexadecimal digit.
    """
    def hex_digit_body(state):
        """ Parse a single hexadecimal digit. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a hexadecimal digit, got the end of input",
                error.HexDigitError.EOI)
        if ('0' <= char <= '9') or ('a' <= char <= 'f') or ('A' <= char <= 'F'):
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected a hexadecimal digit, got '{char}'",
            error.HexDigitError.NOT_DIGIT)
    return hex_digit_body


def newline():
    """
    Return a parser that will match a newline character.

    For Windows users: this will match a single \\r or \\n from a \\n\\r pair.
    """
    def newline_body(state):
        """ Parse a newline character. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a newline, got the end of input",
                error.NewlineError.EOI)
        if ord(char) in _LINE_SEPARATORS:
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected a newline, got '{char}'",
            error.NewlineError.NOT_NEWLINE)
    return newline_body


def nonwhite_char():
    """ Return a parser that will match a character of anything but whitespace. """
    def nonwhite_char_body(state):
        """ Match a non-whitespace character. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a non-whitespace character, got the end of input",
                error.NonwhiteError.EOI)
        if char.isspace():
            raise core.ParsingFailure(
                state,
                "Got a whitespace character when expecting a non-whitespace one",
                error.NonwhiteError.WHITE)
        return state.consume(1)
    return nonwhite_char_body


def white_char(accept_newlines=False):
    """
    Return a parser that will match a character of whitespace, optionally also
    matching newline characters.
    """
    def white_char_body(state):
        """ Match a character of whitespace. """
        try:
            char = state.string[state.left_start]
        except IndexError:
            raise core.ParsingFailure(
                state,
                "Expected a whitespace character, got the end of input",
                error.WhiteCharError.EOI)
        if accept_newlines:
            if char.isspace():
                return state.consume(1)
            raise core.ParsingFailure(
                state,
                f"Expected a whitespace character, got '{char}'",
                error.WhiteCharError.NON_WHITE)
        # not accepting newlines
        if char.isspace():
            if ord(char) in _LINE_SEPARATORS:
                raise core.ParsingFailure(
                    state,
                    f"Got a newline character {hex(ord(char))} when not accepting newlines",
                    error.WhiteCharError.NEWLINE)
            return state.consume(1)
        raise core.ParsingFailure(
            state,
            f"Expected a whitespace character, got '{char}'",
            error.WhiteCharError.NON_WHITE)
    return white_char_body


#--------- aggregates and variations of the above ---------#


def alnum_word(ascii_only=False):
    """
    Return a parser that will match a non-empty sequence of alphanumeric
    character.

    If 'ascii_only' is truthy, match only ASCII characters, otherwise match
    everything that is considered an alphanumeric character in Unicode.
    """
    error_transformer = _mk_aggregate_transformer(
        error.AlnumError.EOI,
        error.AlnumWordError.EOI,
        error.AlnumWordError.NON_ALNUM,
        "Expected an alphanumeric word")
    return core.modify_error(many(alnum(ascii_only), 1), error_transformer)


def alpha_word(ascii_only=False):
    """
    Return a parser that will match a non-empty sequence of alphabetic
    characters.

    If 'ascii_only' is truthy, match only ASCII characters, otherwise match
    everything that is considered an alphabetic character in Unicode.
    """
    error_transformer = _mk_aggregate_transformer(
        error.AlphaError.EOI,
        error.AlphaWordError.EOI,
        error.AlphaWordError.NON_ALPHA,
        "Expected a sequence of alphabetic characters")
    return core.modify_error(many(alpha(ascii_only), 1), error_transformer)


def any_word():
    """
    Return a parser that will match a non-empty sequence of non-whitespace
    characters.
    """
    error_transformer = _mk_aggregate_transformer(
        error.NonwhiteError.EOI,
        error.AnyWordError.EOI,
        error.AnyWordError.WHITE,
        "Expected a sequence of non-whitespace characters")
    return core.modify_error(many(nonwhite_char(), 1), error_transformer)


def hex_int(must_have_prefix=False):
    """
    Return a parser that will match integers in base 16 (with or without '0x'
    prefix).

    If 'must_have_prefix' is truthy, fail if '0x' prefix is omitted.
    """
    def prefix_error_transformer(exc):
        """ Transform the error message about the missing prefix. """
        return core.ParsingFailure(
            exc.state,
            f"Required prefix '0x' is not found in {repr(exc.state.left[:20])}",
            error.HexIntError.NO_PREFIX)
    primary_error_transformer = _mk_aggregate_transformer(
        error.HexDigitError.EOI,
        error.HexIntError.EOI,
        error.HexIntError.NON_HEX_INT,
        "Expected a sequence of hexadecimal digits")
    prefix = literal("0x") if must_have_prefix else maybe(literal("0x"))
    prefix = core.modify_error(prefix, prefix_error_transformer)
    primary = many(hex_digit(), 1)
    primary = core.modify_error(primary, primary_error_transformer)
    return core.chain([prefix, primary])


def integer():
    """
    Return a parser that will match integers in base 10.
    """
    error_transformer = _mk_aggregate_transformer(
        error.DigitError.EOI,
        error.IntegerError.EOI,
        error.IntegerError.NON_INT,
        "Expected a sequence of digits")
    return core.modify_error(many(digit(), 1), error_transformer)


def line(include_newline=False):
    """
    Return a parser that will match a line terminated by a newline.

    If 'include_newline' is truthy, 'parsed' window will contain the
    terminating newline, otherwise it will not.
    """
    def line_body(state):
        """ Match a line optionally terminated by a newline character. """
        pos = 0
        length = state.left_len
        if length == 0:
            raise core.ParsingFailure(
                state,
                "Expected a line, got an end of input",
                error.LineError.EOI)
        while pos < length:
            char = state.string[state.left_start + pos]
            if ord(char) in _LINE_SEPARATORS:
                if include_newline:
                    return state.consume(pos + 1)
                return state._replace(
                    left_start=state.left_start + pos + 1,
                    parsed_start=state.left_start,
                    parsed_end=state.left_start + pos)
            pos += 1
        return state.consume(length)
    return line_body


def whitespace(min_num=1, accept_newlines=False):
    """
    Return a parser that will consume at least 'min_num' whitespace characters,
    optionally with newlines as well.
    """
    def error_transformer(exc):
        """ Transform the error message. """
        if exc.code == error.WhiteCharError.EOI:
            return core.ParsingFailure(
                exc.state,
                "Expected whitespace, got the end of input",
                error.WhitespaceError.EOI)
        if accept_newlines:
            msg = f"Expected at least {min_num} characters of whitespace (and maybe newlines)"
        else:
            msg = f"Expected at least {min_num} characters of whitespace (not newlines)"
        return core.ParsingFailure(
            exc.state,
            f"{msg}, got {repr(exc.state.left[:20])}",
            error.WhitespaceError.NOT_ENOUGH)
    return core.modify_error(many(white_char(accept_newlines), min_num), error_transformer)


#--------- various ---------#


def balanced(opening, closing, include_outer_pair=False):
    """
    Return a parser that will parse everything between an 'opening' string and
    a 'closing' string, where the number of opening and closing strings in the
    parsed part is balanced.

    Fail if the input doesn't start with 'opening' or if there are not enough
    'closing' substrings in the input to balance the opening ones.

    If 'include_outer_pair' is truthy, include the first opening and the last
    closing strings in the 'parsed', otherwise exclude them, leaving just the
    part between them.
    """
    def balanced_body(state):
        """ Match input between balanced pair of strings. """
        open_len = len(opening)
        closing_len = len(closing)
        if state.string[state.left_start:state.left_start + open_len] != opening:
            raise core.ParsingFailure(
                state,
                f"{repr(state.left[0:20])} doesn't start with '{opening}'",
                error.BalancedError.DOESNT_START)
        pos = state.left_start + open_len
        balance = 1
        while pos < state.left_end and balance != 0:
            cur_bit = state.string[pos:state.left_end]
            if cur_bit.startswith(closing):
                balance -= 1
                pos += closing_len
                continue
            if cur_bit.startswith(opening):
                balance += 1
                pos += open_len
                continue
            pos += 1
        if balance != 0:
            raise core.ParsingFailure(
                state,
                f"Failed to find a balanced pair of '{opening}' and '{closing}'",
                error.BalancedError.NO_PAIR)
        if include_outer_pair:
            return state._replace(
                left_start=pos,
                parsed_start=state.left_start,
                parsed_end=pos)
        return state._replace(
            left_start=pos,
            parsed_start=state.left_start + open_len,
            parsed_end=pos - closing_len)
    return balanced_body


def end_of_input():
    """ Return a parser that matches only if there is no input left. """
    def end_of_input_body(state):
        """ Match the end of input. """
        if state.left_start == state.left_end:
            return state._replace()
        raise core.ParsingFailure(
            state,
            f"Expected the end of input, got {repr(state.left[0:20])}",
            error.EndOfInputError.NOT_END)
    return end_of_input_body


def everything():
    """ Return a parser that consumes all remaining input. """
    def everything_body(state):
        """ Consume all remaining input. """
        return state._replace(
            parsed_start=state.left_start,
            parsed_end=state.left_end,
            left_start=state.left_end)
    return everything_body


def literal(lit):
    """
    Return a parser that will match a given literal and remove it from input.
    """
    def literal_body(state):
        """ Match a literal. """
        if state.left_len < len(lit):
            raise core.ParsingFailure(
                state,
                f"{repr(state.left[:20])} doesn't start with {lit}",
                error.LiteralError.SHORTER)
        i = -1
        for i, char in enumerate(lit):
            if char != state.string[state.left_start + i]:
                raise core.ParsingFailure(
                    state,
                    f"'{repr(state.left[:20])} doesn't start with {lit}",
                    error.LiteralError.DOESNT_START)
        return state.consume(i + 1)
    return literal_body


def maybe(parser):
    """
    Return a parser that will match whatever 'parser' matches, and if 'parser'
    fails, matches and consumes nothing.
    """
    def maybe_body(state):
        """
        Match whatever another parser matches, or consume no input if it fails.
        """
        try:
            return parser(state)
        except core.ParsingFailure:
            return state._replace(
                parsed_start=state.left_start,
                parsed_end=state.left_start)
    return core.copy_lookahead(parser, maybe_body)


def many(parser, min_hits=0, max_hits=0, combine=True):
    """
    Return a parser that will run 'parser' on input repeatedly until it fails.

    If 'min_hits' is above zero, fail if 'parser' was run successfully less
    than 'min_hits' times.

    If 'max_hits' is above zero, stop after 'parser' was run successfully
    'max_hits' times.

    If 'combine' is truthy, set 'parsed' of the resulting state object to
    concatenation of individually matched strings, otherwise set it to the last
    matched string.

    Raise ValueError if 'max_hits' is above zero and is less than 'min_hits'.
    """
    if min_hits < 0:
        min_hits = 0
    if max_hits < 0:
        max_hits = 0
    if max_hits > 0 and max_hits < min_hits:
        raise ValueError("'max_hits' is less than 'min_hits'")
    if min_hits > 0:
        must = core.chain(itools.repeat(parser, min_hits), combine)
    else:
        must = None
    if max_hits > 0:
        might = core.chain(itools.repeat(parser, max_hits - min_hits),
                           combine=combine,
                           stop_on_failure=True)
    else:
        might = core.chain(itools.repeat(parser), combine, True)
    if must is None:
        return might
    return core.chain([must, might], combine)


def multi(literals):
    """
    Return a parser that will match any of given literals.
    """
    def error_transformer(exc):
        """ Transform the error message. """
        return core.ParsingFailure(
            exc.state,
            f"None of the literals matched the input: {repr(exc.state.left[:20])}'",
            error.MultiError.ALL_FAILED)
    return core.modify_error(core.branch(map(literal, literals)), error_transformer)


def repeat_while(cond, window_size=1, min_repetitions=0, combine=True):
    """
    Return a parser that will call
    > cond(state, state[:window_size])
    repeatedly consuming 'window_size' characters from the input, until 'cond'
    returns a falsey value. Note that the last window may be less than
    'window_size' characters long.

    If 'min_repetitions' is above 0 and less than that many windows were
    processed, fail.

    If 'combine' is truthy, set 'parsed' of the resulting State object to a
    concatenation of processed windows, otherwise set it to the last window.
    """
    if window_size <= 0:
        raise ValueError("A non-positive 'window_size'")
    def repeat_while_body(state):
        """ Repeatedly check a condition on windows of given width. """
        start = state.left_start
        rep = 0
        while True:
            window_start = state.left_start + rep * window_size
            window_end = window_start + window_size
            window = state.string[window_start:window_end]
            if cond(state, window):
                rep += 1
                continue
            if rep < min_repetitions:
                msg = "Failed to achieve required minimum of repetitions " \
                      f"on input: {repr(state.left[:20])}'"
                raise core.ParsingFailure(state, msg, error.RepeatWhileError.NOT_ENOUGH)
            if combine:
                return state._replace(
                    left_start=window_start,
                    parsed_start=start,
                    parsed_end=window_start)
            return state._replace(
                left_start=window_start,
                parsed_start=window_start - window_size,
                parsed_end=window_start)
    return repeat_while_body


def take(num, fail_on_fewer=True):
    """
    Return a parser that will consume exactly 'num' characters.

    If 'fail_on_fewer' is truthy, fail if fewer than 'num' characters are
    available.

    Raise ValueError if 'num' is negative.
    """
    if num < 0:
        raise ValueError("Negative number of consumed characters")
    def take_body(state):
        """ Consume a fixed number of characters. """
        if fail_on_fewer and state.left_len < num:
            msg = "Less than requested number of characters received on input: " \
                  f"{repr(state.left[:20])}'"
            raise core.ParsingFailure(state, msg, error.TakeError.NOT_ENOUGH)
        return state.consume(min(num, state.left_len))
    return take_body


def weave(parsers, separator, trailing=None, stop_on_failure=False):
    """
    Return a chain where each parser in 'parsers' is separated by 'separator'
    from others.

    If 'trailing' is not None, append it to the resulting chain.

    'stop_on_failure' will be sent through to the underlying 'chain' generator.
    """
    parsers = iter(parsers)
    saved = deque()
    def iterator():
        """ Return the iterable that will go into the chain. """
        saved_len = len(saved)
        for i, p in enumerate(itools.chain(saved, parsers)):
            if i != 0:
                yield separator
            if i >= saved_len:
                saved.append(p)
            yield p
        if trailing is not None:
            yield trailing
    return core.chain(core.reuse_iter(iterator), stop_on_failure=stop_on_failure)


#--------- helper things ---------#


_LINE_SEPARATORS = [0x000a, 0x000d, 0x001c, 0x001d, 0x001e, 0x0085, 0x2028, 0x2029]


def _mk_aggregate_transformer(
        eoi_code_in,
        eoi_code_out,
        generic_code_out,
        msg):
    """ Create an error transformer for aggregate functions. """
    def transformer(exc):
        """ Transform the error message. """
        if exc.code == eoi_code_in:
            return core.ParsingFailure(
                exc.state,
                f"{msg}, got the end of input",
                eoi_code_out)
        return core.ParsingFailure(
            exc.state,
            f"{msg}, got {repr(exc.state.left[:20])}",
            generic_code_out)
    return transformer
